/*
 * bulkthread.c
 * sound playing thread program using USB bulk transfer
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <semaphore.h>
#include <locale.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <asm/byteorder.h>

#include "common.h"
#include "wav.h"
#include "usbmydev.h"


#ifndef	WITHOUT_DSD
#define	MUTE_OFF_COUNT	4096
static const char dev_xor[] = { 0x16, 0x16, 0x96, 0x96 };
static const unsigned char dev_zeropoint[] = { 0x80, 0x80, 0x00 };
static const unsigned char dop1_zeropoint[] = {
	0x96, 0x96, 5, 0x96, 0x96, 5, 0x96, 0x96, 0xfa, 0x96, 0x96, 0xfa };
static const unsigned char dop2_zeropoint[] = {
	0x96, 0x96, 6, 0x96, 0x96, 6, 0x96, 0x96, 0xf9, 0x96, 0x96, 0xf9 };
#endif	/* WITHOUT_DSD */

#define PERIOD_TIME 100000	/* 100,000 usec == 0.1 sec*/
#define BUFFER_TIME (PERIOD_TIME * 4)
#define WAIT_TIME 5000
#define WAIT_SHORT 1000

extern MusicFile *musicfile;
static MusicFile *p_dumprequest;
extern MusicFile *p_current;
extern int musicfile_current;
extern int files_open;

#define BUFFER_MARK_NONE		0
#define BUFFER_MARK_BEFORECHANGE	1
#define BUFFER_MARK_AFTERCHANGE		2
#define BUFFER_MARK_AFTERFORMATCHANGE	4
extern SoundBuffer *soundbuffer;
extern int buffer_num;
extern int soundbuffer_towrite;
extern int soundbuffer_filled;

extern sem_t semaphore;
extern sem_t callback_semaphore;

extern int divider;

#define OPEN_RETRY_TIME 4
extern char devname[];

static int request_formatchange;

#define	DEFAULT_PLAY_WAIT 50
static int play_wait;

enum e_format {E_START, E_DSD, E_PCM};
#ifndef	WITHOUT_DSD
static enum e_format e_format_remember = E_START;
static int mute_off_count;
#endif
char xor;
unsigned char zeropoint;

int thread_end_flag = FALSE;
E_DSD_TYPE	e_dsd_type = E_CQ;
extern E_DSD_TYPE	e_dsd_playback;

static int usb_fd;
static int submit_count = 0;

#ifndef	WITHOUT_DSD
static inline void fill_silent(AudioSpec *pas, char *stream, int channels, int to_fill)
{
  memset(stream, 0x96, to_fill);
}
#endif	/* WITHOUT_DSD*/

int sound_push()
{
  int ret;
  int status = 1;
  SoundBuffer *psb = &soundbuffer[soundbuffer_towrite];
  AudioSpec *pas = &psb->pmf->fmt;
  int to_fill, fill_count;
  static int read_start = 0;
  int read_left;
  char stream[512];
  struct usbdevfs_urb send, receive, *p_receive;


  DEBUG_PRINT("I");

  memset(&send, 0, sizeof (send));
  send.buffer_length = 512;
  send.buffer = stream;
  send.type = USBDEVFS_URB_TYPE_BULK;
  send.endpoint = 0x03;
  send.signr = 0;

#if 1
  while (submit_count > 0 && usb_reapurb_ndelay(usb_fd, &p_receive) >= 0) {
    submit_count--;
  }
  if (submit_count > 6) {
    usleep(1000);
    goto finish;
  }
#endif
  to_fill = 504;
#ifdef	WITHOUT_DSD
  memset(stream, 0, to_fill);
#else	/* WITHOUT_DSD */
  fill_silent(pas, stream, 2, to_fill);
  stream[511] = to_fill / 6;
#endif
  if (play_wait <= 0 && soundbuffer_filled <= 0 ) {
    fill_count = 0;
    status = 0;
    goto finish;
  }

  fill_count = 0;
#ifndef	WITHOUT_DSD
  /* UDA3 mute off */
  if ( mute_off_count > 0 ) {
    if ( to_fill > mute_off_count ) {
      mute_off_count = 0;
    } else {
      mute_off_count -= to_fill;
    }
    if (play_wait > 0) {
      play_wait--;
    }
    if (usb_submiturb(usb_fd, &send) < 0) {
      perror("");
#if 0
      exit (1);
#endif
    } else {
      //usb_reapurb(usb_fd, &receive);
      submit_count++;
    }
    return (1);
  }
#endif	/* WITHOUT_DSD */

  if (play_wait > 0) {
    play_wait--;
    if (usb_submiturb(usb_fd, &send) < 0) {
      perror("");
#if 0
      exit (1);
#endif
    } else {
      submit_count++;
      //usb_reapurb(usb_fd, &receive);
    }
    return (1);
  }

  psb = &soundbuffer[soundbuffer_towrite];
  read_left = psb->buffer_start - read_start;

  while ( to_fill > 0 ) {
    while ( read_left <= 0 ) {
      // get data from next buffer
      if ( soundbuffer_filled <= 0 ) {
	play_wait = DEFAULT_PLAY_WAIT;
	read_start = read_left = 0;
	goto finish;
      }
      sem_wait(&semaphore);
      soundbuffer_filled--;
      sem_post(&semaphore);
      if ( soundbuffer_filled <= 0 ) {
	play_wait = DEFAULT_PLAY_WAIT;
	read_start = read_left = 0;
	goto finish;
      }
      psb->pmf->left_buffer_count--;
      CYCLIC_INCREMENT(soundbuffer_towrite, buffer_num);
      psb = &soundbuffer[soundbuffer_towrite];
      if ( (psb->buffer_mark & BUFFER_MARK_AFTERCHANGE) != 0 ) {
        p_dumprequest = psb->pmf;
        if ( (psb->buffer_mark & BUFFER_MARK_AFTERFORMATCHANGE) != 0 ) {
          request_formatchange = TRUE;
          psb->buffer_mark = 0;
          status = 0;
          goto finish;
        }
      }
      read_start = 0;
      read_left = psb->buffer_start;
    }

    // write data from buffer to device
    int size = psb->buffer_start - read_start;
    if ( size > to_fill ) {
      size = to_fill;
    }
    DEBUG_ASSERT(psb->buffer_start <= BUFFER_SIZE);
    DEBUG_ASSERT(psb->buffer_start % pas->bytes_per_outframe == 0);
    DEBUG_ASSERT(read_start % pas->bytes_per_outframe == 0);
    DEBUG_ASSERT(to_fill % pas->bytes_per_outframe == 0);
    memcpy(stream, psb->audio_buf + read_start, size);
    stream[511] = size / pas->bytes_per_outframe;
    if (usb_submiturb(usb_fd, &send) < 0) {
      perror("");
#if 0
      exit (1);
#else
      goto finish;
#endif
    } else {
      submit_count++;
      //usb_reapurb(usb_fd, &receive);
    }
    read_start += size;
    read_left -= size;
    to_fill -= size;
    fill_count += size / pas->bytes_per_outframe;
    psb->pmf->sample_output += size / pas->bytes_per_outframe * 2;
  }

  DEBUG_PRINT("p%d", soundbuffer_filled);
  if (fill_count == 0) {
    DEBUG_PRINT("I0");
    return (status);
  }

finish:
  DEBUG_PRINT("Io");
  return (status);
}

void	sound_format_test(MusicFile *pmf)
{
    pmf->fmt.bytes_per_insample
      = snd_pcm_format_physical_width(pmf->fmt.input_format) / 8;
    pmf->fmt.bytes_per_inframe = pmf->wfx.Format.nChannels
      * pmf->fmt.bytes_per_insample;

    pmf->fmt.output_format = SND_PCM_FORMAT_S24_3LE;
    pmf->fmt.bytes_per_outsample = 3;
    pmf->fmt.bytes_per_outframe = 6;
}

void	sound_init(MusicFile *pmf)
{
	char buff[512];
	int size;
	int frequency;
	struct usbdevfs_urb send, receive;
	char *devname;

	devname = rad001_getdevname();
	if (devname == NULL) {
		fprintf(stderr, "device not found.\n");
		exit (1);
	}

	usb_fd = usb_open(devname);
	usb_setinterface(usb_fd, 0, 0);

        memset(buff, 0, sizeof(buff));
	buff[0] = 'P';
	memset(&send, 0, sizeof (send));
	send.buffer_length = 512;
	send.buffer = buff;
	send.type = USBDEVFS_URB_TYPE_BULK;
	send.endpoint = 0x01;
	send.signr = 0;
        switch(pmf->fmt.p_wfx->Format.nSamplesPerSec) {
	case 117600:
	case 235200:
	case 470400:
		frequency = pmf->fmt.p_wfx->Format.nSamplesPerSec * 24;
		break;
	default:
		frequency = pmf->fmt.p_wfx->Format.nSamplesPerSec;
		break;
	}
        memcpy(buff + 4, &frequency, sizeof(int));
        if ( usb_submiturb(usb_fd, &send) < 0) {
                perror("error on set frequency");
		exit(1);
	} else if ( usb_reapurb(usb_fd, &receive) < 0 ) {
		perror("");
		exit(1);
	}
	usleep(100000);
}

void    sound_close(int change_format)
{
        close(usb_fd);
}

#ifndef	WITHOUT_DSD
void	select_device(void);
void	sound_get_dsd_type()
{
  e_dsd_type = E_CQ;
}
#endif	/* WITHOUT_DSD */

void soundDisplay(MusicFile *pmf)
{
  static time_t oldtm = 0;
  time_t tm;
  int hour, min, sec;
  int buffer_fill;

  if (play_wait > 0) {
    int i;
    printf(CR "waiting ");
    for ( i = 0; i < 10*(DEFAULT_PLAY_WAIT - play_wait)/DEFAULT_PLAY_WAIT; i++ ) {
      printf("=");
    }
    for ( ; i < 10; i++ ) {
      printf(".");
    }
    fflush(stdout);
    return;
  }

  if ( pmf == NULL ) {
    return;
  }

  time(&tm);
  if ( tm == oldtm ) {
    return;
  }

  oldtm = tm;

  sec = pmf->sample_output / pmf->wfx.Format.nSamplesPerSec / pmf->wfx.Format.nChannels;
  min = sec / 60;
  hour = min / 60;
  min %= 60;
  sec %= 60;
  buffer_fill = soundbuffer_filled + 1;
  if ( buffer_fill > buffer_num ) {
    buffer_fill = buffer_num;
  }
  if (soundbuffer_filled <= 0) {
    printf(CR "                                                      ");
  } else {
    printf(CR "playing position %d:%02d:%02d buffer %d/%d            ",
	   hour, min, sec, buffer_fill, buffer_num);
  }

  fflush(stdout);
}

void *playloop(void *dummy)
{
  int initialized = FALSE;
  unsigned char *p_fileshortname;
#ifndef	WITHOUT_DSD
  enum e_format e_format_new;
#endif

  thread_end_flag = FALSE;

  /* initialize sound */
  p_current = NULL;
  request_formatchange = TRUE;
  p_dumprequest = musicfile;

  /* main play loop */
  while ( soundbuffer_filled > 0 ) {
    if ( !request_formatchange && p_dumprequest != p_current ) {
      printf(CR "                                                     \n"
	     "playing file %d/%d ", musicfile_current + 1, files_open);
      int sec;
      if (p_dumprequest->wfx.Format.nSamplesPerSec > 96000) {
      	sec = p_dumprequest->total_len / p_dumprequest->wfx.Format.nSamplesPerSec / 64;
      } else {
      	sec = p_dumprequest->total_len / p_dumprequest->wfx.Format.nSamplesPerSec / p_dumprequest->wfx.Format.nBlockAlign;
      }
      int min = sec / 60;
      int hour = min / 60;
      min %= 60;
      sec %= 60;
      p_fileshortname = (unsigned char *)rindex(p_dumprequest->filename, '/');
      if (p_fileshortname == NULL) {
        printf("total %d:%02d:%02d \"%s\"\n", hour, min, sec,
		p_dumprequest->filename);
      } else {
        printf("total %d:%02d:%02d \"%s\" \"%s\"\n", hour, min, sec,
		p_fileshortname + 1, p_dumprequest->filename);
      }

      musicfile_current++;
      p_current = p_dumprequest;
    }
    if ( request_formatchange ) {
#ifdef	WITHOUT_DSD
      if ( initialized ) {
        sound_close(FALSE);
      }
#else
      e_format_new = (soundbuffer[soundbuffer_towrite].pmf->file_type == E_WAV) ? E_PCM : E_DSD;
      if ( initialized) {
        sound_close((e_format_new != e_format_remember));
      }
#endif
      sound_init(soundbuffer[soundbuffer_towrite].pmf);
      initialized = TRUE;
      request_formatchange = FALSE;
    }
    sound_push();
    //soundDisplay(p_current);
    usleep(60);

  }
  printf(CR "                                                 \n");
  fflush(stdout);

  sound_close(TRUE);

  thread_end_flag = TRUE;
  return NULL;
}
