/*
 * udaplay.c
 * sound playing program using ALSA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <semaphore.h>
#include <locale.h>
#include <pthread.h>
#include <sys/types.h>
#include <signal.h>
#include <asm/byteorder.h>

#include "common.h"
#include "wav.h"
#include "dir2files.h"
#ifdef WITHOUT_DSD
#include "ffmpeg.h"
#else
#include "dsd.h"
#include "wsd.h"
#endif	/* WITHOUT_DSD */
#ifdef	DSD22M
#include "coeff44k352k.h"
#endif

#define WAVE_FORMAT_EXTENSIBLE	0xfffe

#define WAIT_TIME 5000
#define WAIT_SHORT 1000

static int count_mode = FALSE;
static int counted_files = 0;

MusicFile *musicfile = NULL;
MusicFile *p_current;
static MusicFile *p_reading;
int musicfile_current = 0;
static int musicfile_reading = 0;
int files_open;

SoundBuffer *soundbuffer = NULL;
int buffer_num;
static int soundbuffer_toread = 0;
int soundbuffer_towrite = 0;
volatile int soundbuffer_filled = 0;
int divider = 1;

sem_t semaphore;
sem_t callback_semaphore;

#define OPEN_RETRY_TIME 4
char devname[1024] = {0};

static int skip_sec = 0;
#ifndef	WITHOUT_DSD
static int DoPcount1 = 0;
static int DoPcount2 = 0;
#endif
static SoundBuffer *psb_prev = NULL;
static int interrupted = FALSE;

static const char *copyright = "Programmed & Copyright (c) 2010,2011,2012,2013,2014 by NAKATA Hiroshi <webmaster@nakata-jp.org>. All rights reserved."
#ifdef	WITHOUT_DSD
#ifdef	FFMPEG
"\nPowered by ffmpeg library (LGPL ver.2.1)."
#endif
#else
"\nCollaborated with ElectrArt"
#endif
;

extern void	sound_format_test(MusicFile *pmf);
extern void	select_device(void);
extern void	*playloop(void *);
#ifndef	WITHOUT_DSD
extern void	sound_get_dsd_type();
extern E_DSD_TYPE	e_dsd_type;
#endif
extern unsigned char	zeropoint;
extern char	xor;
extern int	thread_end_flag;
E_DSD_TYPE	e_dsd_playback;

extern int myseek(int fd, off_t offset, int whence);

static int     sounddata_closefile(MusicFile *pmf)
{
    if (pmf->fd > 0) {
    	close(pmf->fd);
	pmf->fd = -1;

#ifdef	FFMPEG
        closeFfmpeg(pmf);
#endif
    }

    return (0);
}

static int     sounddata_openfile(MusicFile *pmf)
{
  pmf->fd = open(pmf->filename, O_RDONLY);
  if ( pmf->fd < 0 ) {
    fprintf(stderr, "file %s open error.\n", pmf->filename);
    exit (1);
  }

  if ( readWavHeader(pmf) == 0 ) {
    pmf->file_type = E_WAV;
    e_dsd_playback = E_NOTDSD;
    goto wav_opend;
  }
#ifdef	FFMPEG
  if ( readFfmpegHeader(pmf) == 0 ) {
    pmf->file_type = E_FFMPEG;
    goto wav_opend;
  }
  fprintf(stderr, "file format of %s unknown.\n", pmf->filename);
  exit (1);
#else
  myseek(pmf->fd, 0, SEEK_SET);
  if ( readDsdHeader(pmf) == 0 ) {
    pmf->file_type = E_DSDIFF;
    goto dsd_opend;
  }

  myseek(pmf->fd, 0, SEEK_SET);
  if ( readWsdHeader(pmf) == 0 ) {
    pmf->file_type = E_WSD;
    goto dsd_opend;
  }
  if (count_mode) {
    return -1;
  }
  fprintf(stderr, "file %s is not WAV,DeltaSigma format.\n", pmf->filename);
  exit (1);


 dsd_opend:
  if (count_mode) {
    return 0;
  }
  e_dsd_playback = e_dsd_type;
  pmf->total_len *= 32;
  pmf->total_len /= 3;
    pmf->wfx.Format.nChannels = 2;
    // CQ format
    pmf->wfx.Format.nSamplesPerSec *= 4;
    pmf->wfx.Format.nSamplesPerSec /= 3;
    pmf->wfx.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;
    pmf->wfx.Format.nAvgBytesPerSec = pmf->wfx.Format.nSamplesPerSec * 6;
    pmf->wfx.Format.nBlockAlign = 6;
    pmf->wfx.Format.wBitsPerSample = 24;
    pmf->wfx.Format.cbSize = 22;
    pmf->wfx.Samples.wValidBitsPerSample = pmf->wfx.Format.wBitsPerSample;
    pmf->wfx.dwChannelMask = 0;
#endif	/* WITHOUT_DSD */

 wav_opend:
  if (count_mode) {
    return 0;
  }
  pmf->sound_left_flag = TRUE;
  pmf->left_buffer_count = 0;
  pmf->fmt.p_wfx = &pmf->wfx;
#ifdef	FFMPEG
  //if (pmf->file_type != E_FFMPEG) {
#endif
  switch ( pmf->wfx.Format.wBitsPerSample ) {
  case 8:
#ifdef	WITHOUT_DSD
    pmf->fmt.input_format = SND_PCM_FORMAT_U8;
#else	/* WITHOUT_DSD */
    if ( zeropoint == 0x80 ) {
      pmf->fmt.input_format = SND_PCM_FORMAT_U8;
    } else {
      pmf->fmt.input_format = SND_PCM_FORMAT_S8;
    }
#endif	/* WITHOUT_DSD */
    break;
  case 16:
    pmf->fmt.input_format = SND_PCM_FORMAT_S16_LE;
    break;
  case 24:
    if ( pmf->wfx.Format.nBlockAlign / pmf->wfx.Format.nChannels == 4 ) {
      pmf->fmt.input_format = SND_PCM_FORMAT_S24_LE;
    } else {
      pmf->fmt.input_format = SND_PCM_FORMAT_S24_3LE;
    }
    break;
  case 32:
    if ( pmf->wfx.Format.wFormatTag == 3 ||
	 pmf->wfx.SubFormat == 3 ) {
      pmf->fmt.input_format = SND_PCM_FORMAT_FLOAT_LE;
    } else {
      pmf->fmt.input_format = SND_PCM_FORMAT_S32_LE;
    }
    break;
  default:
    exit (1);
  }
#ifdef	FFMPEG
  //}
#endif
#if 1
  if ( pmf->wfx.Format.nSamplesPerSec > 470400 ) {
    fprintf(stderr, "sampling rate %dHz is too large\n", pmf->wfx.Format.nSamplesPerSec * 24);
    exit(1);
  }
  if ( pmf->wfx.Format.nSamplesPerSec != 117600 &&
       pmf->wfx.Format.nSamplesPerSec != 235200 &&
       pmf->wfx.Format.nSamplesPerSec != 470400 &&
       pmf->wfx.Format.nSamplesPerSec > 96000 ) {
    fprintf(stderr, "sampling rate %dHz is too large\n", pmf->wfx.Format.nSamplesPerSec);
    exit(1);
  }
#endif
#ifdef	DSD22M
  if ( pmf->wfx.Format.nSamplesPerSec != 44100) {
    fprintf(stderr, "sampling rate %dHz is not supported yet\n", pmf->wfx.Format.nSamplesPerSec);
    exit(1);
  }
#endif
  return (0);
}

#ifndef	DSD22M
static int simple_read_audio(MusicFile *pmf)
{
  int size, result = 0;
  SoundBuffer *psb = &soundbuffer[soundbuffer_toread];
#ifndef	WITHOUT_DSD
  unsigned char *p;
  int i;
#endif

  while ( psb->buffer_start <= BUFFER_SIZE - pmf->wfx.Format.nBlockAlign ) {
    if ( pmf->data_len <= 0 ) {
      pmf->sound_left_flag = FALSE;
      if ( psb->buffer_start > 0 ) {
	psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	goto flush_buffer;      /* flush last buffer of previous sound */
      } else if ( psb_prev != NULL ) {
	psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
      }
    }
    size = BUFFER_SIZE - psb->buffer_start;
    if ( pmf->data_len < size ) {
      size = pmf->data_len;
    }

    unsigned char * const local_buff = (unsigned char*)psb->audio_buf + psb->buffer_start;
    if (result <= 0) {
      switch (pmf->file_type) {
      case E_WAV:
        result = readWavData(pmf, local_buff, size);
        break;
#ifdef	FFMPEG
      case E_FFMPEG:
	DEBUG_ASSERT(0);
#endif
#ifndef	WITHOUT_DSD
      case E_DSDIFF:
        result = readDsdData(pmf, local_buff, size);
        break;
      case E_WSD:
        result = readWsdData(pmf, local_buff, size);
        break;
#endif
      }
    }
#ifndef	WITHOUT_DSD
    if (pmf->file_type != E_WAV) {
      int i;
      switch (e_dsd_playback) {
      case E_DOP_1:
	for (i = 0; i < result; i += 6) {
	  local_buff[i + 2] = local_buff[i + 5] = (DoPcount2 == 0) ? 5 : 0xfa;
	  DoPcount2 = (DoPcount2 + 1) % 2;
	}
	break;
      case E_DOP_2:
	for (i = 0; i < result; i += 3) {
	  local_buff[i + 2] = (DoPcount2 == 0) ? 6 : 0xf9;
	  if (++DoPcount1 >= pmf->wfx.Format.nSamplesPerSec / 88200 * 2) {
	    DoPcount1 = 0;
	    DoPcount2 = (DoPcount2 + 1) % 2;
	  }
	}
	break;
      default:
	break;
      }
    }
#endif
    if ( result <= 0 ) {
      pmf->sound_left_flag = FALSE;
      pmf->data_len = 0;
      DEBUG_PRINT("e1");
      if ( psb->buffer_start > 0 ) {
	psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	goto flush_buffer;      /* flush last buffer of previous sound */
      }
      if ( psb_prev != NULL ) {
	psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
      }
      DEBUG_PRINT("Rse");
      return (psb->buffer_start);
    }
    psb->buffer_start += result;
#ifdef	FFMPEG
    if ( pmf->file_type != E_FFMPEG ) {
      pmf->data_len -= result;
    }
#else
    pmf->data_len -= result;
#endif
    DEBUG_PRINT("r");
  }
flush_buffer:
  if ( psb_prev != NULL && psb_prev->pmf != psb->pmf ) {
    sounddata_closefile(psb_prev->pmf);
  }
  if ( psb->buffer_start > 0 ) {
    psb_prev = psb;
    CYCLIC_INCREMENT(soundbuffer_toread, buffer_num);
    memset(&soundbuffer[soundbuffer_toread], 0, sizeof(SoundBuffer));
    sem_wait(&semaphore);
    pmf->left_buffer_count++;
    soundbuffer_filled++;
    sem_post(&semaphore);

    return (psb->buffer_start);
  }

  return (psb_prev->buffer_start);
}
#endif	// DSD22M

static int read_audio(MusicFile *pmf)
{
  static char local_buf[1536];
#ifndef	WITHOUT_DSD
  int i;
#endif
  int size;
  int read_count = 0;
  SoundBuffer *psb = &soundbuffer[soundbuffer_toread];
  static int result = -1;
  static int local_buf_start = 0;
  static int local_buf_len = 0;

  if ( soundbuffer_filled >= (buffer_num - 2) || !pmf->sound_left_flag ) {
    return (0);          /* means no error */
  }

  DEBUG_PRINT("Rs");
  memset(psb, 0, sizeof(SoundBuffer));
  psb->pmf = pmf;
  if ( psb_prev == NULL ) {
#if 0
    psb->buffer_mark |= BUFFER_MARK_AFTERCHANGE;
    psb->buffer_mark |= BUFFER_MARK_AFTERFORMATCHANGE;
#endif
  } else if ( psb_prev->pmf != pmf ) {
    psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
    psb->buffer_mark |= BUFFER_MARK_AFTERCHANGE;
    if ( psb_prev->pmf->wfx.Format.nSamplesPerSec != psb->pmf->wfx.Format.nSamplesPerSec
	 || psb_prev->pmf->wfx.Format.nBlockAlign != psb->pmf->wfx.Format.nBlockAlign
	 || psb_prev->pmf->wfx.Format.nChannels != psb->pmf->wfx.Format.nChannels ) {
      psb->buffer_mark |= BUFFER_MARK_AFTERFORMATCHANGE;
    }
  }
  const AudioSpec *pas = &(psb->pmf->fmt);
#ifndef	DSD22M
  if ( pas->output_format == pas->input_format
#ifdef	FFMPEG
 	&& pmf->file_type != E_FFMPEG
#endif
      ) {
    return (simple_read_audio(pmf));
  }
#endif	//DSD22M

  while ( (psb->buffer_start <= BUFFER_SIZE - pmf->wfx.Format.nBlockAlign) ||
	  (psb->buffer_start % pmf->wfx.Format.nBlockAlign != 0) ) {

    if ( local_buf_len <= 0 ) {
      if ( pmf->data_len <= 0 ) {
	pmf->sound_left_flag = FALSE;
	if ( psb->buffer_start <= 0 ) {
	  if ( psb_prev != NULL ) {
	    psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	  }
	} else {
	  psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
        }
	goto flush_buffer;      /* flush last buffer of previous sound */
      }
      size = sizeof(local_buf);
#ifndef	FFMPEG
      if ( pmf->data_len < size ) {
	size = pmf->data_len;
      }
#endif
      size -= (size % pmf->wfx.Format.nBlockAlign);
      
      switch (pmf->file_type) {
      case E_WAV:
	result = readWavData(pmf, (unsigned char *)local_buf, size);
	break;
#ifdef	FFMPEG
      case E_FFMPEG:
	while ( (result = readFfmpegData(pmf, (unsigned char *)local_buf, sizeof(local_buf))) == 0)
		;	// empty loop body
	break;
#else  /* WITHOUT_DSD */
#if 0
      case E_DSDIFF:
	result = readDsdData(pmf, (unsigned char *)local_buf, size, IS_DOP(e_dsd_type));
	break;
      case E_WSD:
	result = readWsdData(pmf, (unsigned char *)local_buf, size, IS_DOP(e_dsd_type));
	break;
#else
      default:
	break;
#endif
#endif	/* WITHOUT_DSD */
      }
      DEBUG_ASSERT(result <= (int)sizeof(local_buf));
#ifndef	WITHOUT_DSD
      if (pmf->file_type != E_WAV) {
	switch (e_dsd_playback) {
	case E_DOP_1:
	  for (i = 0; i < result; i += 6) {
	    local_buf[i + 2] = local_buf[i + 5] = (DoPcount2 == 0) ? 5 : 0xfa;
	    DoPcount2 = (DoPcount2 + 1) % 2;
	  }
	  break;
	case E_DOP_2:
	  for (i = 0; i < result; i += 3) {
	    local_buf[i + 2] = (DoPcount2 == 0) ? 6 : 0xf9;
	    if (++DoPcount1 >= pmf->wfx.Format.nSamplesPerSec / 88200 * 2) {
	      DoPcount1 = 0;
	      DoPcount2 = (DoPcount2 + 1) % 2;
	    }
	  }
	  break;
	default:
	  break;
	}
      }
#endif	/* WITHOUT_DSD */
      if ( result <= 0 ) {
#ifndef	WITHOUT_DSD
	if (e_dsd_playback == E_DOP_2 && DoPcount1 != 0) {
	  for (i = 0; i < result; i += 3) {
	    local_buf[i] = local_buf[i + 1] = 0x96;
	    local_buf[i + 2] = (DoPcount2 == 0) ? 6 : 0xf9;
	    if (++DoPcount1 >= pmf->wfx.Format.nSamplesPerSec / 88200 * 2) {
	      DoPcount1 = 0;
	      DoPcount2 = (DoPcount2 + 1) % 2;
	    }
	  }
	}
#endif	/* WITHOUT_DSD */
	pmf->sound_left_flag = FALSE;
	pmf->data_len = 0;
	local_buf_len = 0;
	local_buf_start = 0;
	DEBUG_PRINT("e2");
	if ( psb->buffer_start > 0 ) {
	  psb->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	  //sounddata_closefile(psb->pmf);
	} else if ( psb_prev != NULL ) {
	  sounddata_closefile(psb_prev->pmf);
	  psb_prev->buffer_mark |= BUFFER_MARK_BEFORECHANGE;
	}
	goto flush_buffer;
      }
      read_count += result;
      local_buf_len = result;
      local_buf_start = 0;
#ifdef	FFMPEG
      if ( pmf->file_type != E_FFMPEG ) {
        pmf->data_len -= result;
      }
#else
      pmf->data_len -= result;
#endif
      DEBUG_PRINT("Re");
    }

    union {
      long l;
      float f;
    } u;
    int32_t l;
    float f;
    double d;
    unsigned char *p_src;
    unsigned char *p_dst;
    unsigned char *p_data;

    p_dst = &(psb->audio_buf[psb->buffer_start]);
    p_src = (unsigned char *)&local_buf[local_buf_start];

    while (local_buf_len > 0 && psb->buffer_start < BUFFER_SIZE) {
      p_data = (unsigned char *)&l;
      switch ( pas->input_format ) {
      case SND_PCM_FORMAT_U8:
	l = *p_src;
	l -= 0x80;
	l <<= 24;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S16_LE:
	l = __le16_to_cpu(*(__le16 *)p_src);
	l <<= 16;
	d = l / (double)0x80000000UL;
	break;
      case SND_PCM_FORMAT_U16_LE:
	l = __le16_to_cpu(*(__le16 *)p_src);
	l <<= 16;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S16_BE:
	l = __be16_to_cpu(*(__be16 *)p_src);
	l <<= 16;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U16_BE:
	l = __be16_to_cpu(*(__be16 *)p_src);
	l <<= 16;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S32_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U32_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S32_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U32_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	l <<= 8;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_LE:
	l = __le32_to_cpu(*(__le32 *)p_src);
	l <<= 8;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	l <<= 8;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_BE:
	l = __be32_to_cpu(*(__be32 *)p_src);
	l <<= 8;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S24_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U24_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S18_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 6;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U18_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 6;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S18_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 6;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U18_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 6;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S20_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 4;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U20_3LE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[0];
	p_data[2] = p_src[1];
	p_data[3] = p_src[2];
#else
	p_data[3] = 0;
	p_data[2] = p_src[0];
	p_data[1] = p_src[1];
	p_data[0] = p_src[2];
#endif
	l <<= 4;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_S20_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 4;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_U20_3BE:
#ifdef	__LITTLE_ENDIAN
	p_data[0] = 0;
	p_data[1] = p_src[2];
	p_data[2] = p_src[1];
	p_data[3] = p_src[0];
#else
	p_data[3] = 0;
	p_data[2] = p_src[2];
	p_data[1] = p_src[1];
	p_data[0] = p_src[0];
#endif
	l <<= 4;
	l -= 0x80000000;
	d = l / (double)0x80000000;
	break;
      case SND_PCM_FORMAT_FLOAT_LE:
	u.l = __le32_to_cpu(*(__le32 *)p_src);
	f = u.f;
	l = f * 0x80000000;
	d = (double)f;
	break;
      case SND_PCM_FORMAT_FLOAT_BE:
	u.l = __be32_to_cpu(*(__be32 *)p_src);
	f = u.f;
	l = f * 0x80000000;
	d = (double)f;
	break;
      case SND_PCM_FORMAT_MU_LAW:
	fprintf(stderr,
		"This program does'nt support Mu-Law format.\n");
	exit(1);
      default:
	fprintf(stderr, "Unsupported input format %d.\n",
		psb->pmf->fmt.input_format);
	exit(1);
      }
      p_src += pas->bytes_per_insample;
      f = (float)d;

#ifdef	DSD22M
      static int left_flag = TRUE;
      static unsigned char *left_channel, *right_channel;
      int i;

      local_buf_len -= pas->bytes_per_insample;
      local_buf_start += pas->bytes_per_insample;
      if (left_flag) {
        left_flag = !left_flag;
	left_channel = coeff44k352k22M_convert(f, 0);
	DEBUG_ASSERT(left_channel != NULL);
      } else {
        left_flag = !left_flag;
	right_channel = coeff44k352k22M_convert(f, 1);
	DEBUG_ASSERT(right_channel != NULL);
	int buflen = IO_FREQUENCY_PROPORTION * IO_FREQUENCY_PROPORTION2 / BITS_IN_BYTE / divider;
	if (divider == 1) {
	  for ( i = 0; i < buflen; i++) {
	    *p_dst++ = left_channel[i];
	    *p_dst++ = right_channel[i];
	  }
          psb->buffer_start += (buflen * 2);
	} else {
	  for (i = 0; i < buflen; i += 4) {
	    *p_dst++ = left_channel[i+1];
	    *p_dst++ = left_channel[i];
	    *p_dst++ = 5;
	    *p_dst++ = right_channel[i+1];
	    *p_dst++ = right_channel[i];
	    *p_dst++ = 5;
	    *p_dst++ = left_channel[i+3];
	    *p_dst++ = left_channel[i+2];
	    *p_dst++ = 0xfa;
	    *p_dst++ = right_channel[i+3];
	    *p_dst++ = right_channel[i+2];
	    *p_dst++ = 0xfa;
	  }
	  psb->buffer_start += (buflen * 3);
	}
	if (psb->buffer_start >= BUFFER_SIZE) {
	  goto flush_buffer;
	}
      }
#else	// DSD22M
      switch ( pas->output_format ) {
      case SND_PCM_FORMAT_S16_LE:
	p_data += 2;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S24_LE:
	l >>= 8;
	if ( l & 0x800000 ) {
	  l |= 0xff000000;
	}
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S32_LE:
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_FLOAT_LE:
	memcpy(p_dst, &f, 4);
	p_dst += 4;
	break;
      case SND_PCM_FORMAT_FLOAT64_LE:
	memcpy(p_dst, &d, 8);
	p_dst += 8;
	break;
      case SND_PCM_FORMAT_S24_3LE:
	p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S20_3LE:
	l >>= 4;
	if ( l & 0x8000000 ) {
	  l |= 0xf0000000;
	}
	p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S18_3LE:
	l >>= 6;
	if ( l & 0x4000000 ) {
	  l |= 0xfc000000;
	}
	p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	*p_dst++ = *p_data++;
	break;
      case SND_PCM_FORMAT_S16_BE:
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	break;
      case SND_PCM_FORMAT_S24_BE:
	l >>= 8;
	if ( l & 0x800000 ) {
	  l |= 0xff000000;
	}
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_S32_BE:
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_S24_3BE:
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_S20_3BE:
	l >>= 4;
	if ( l & 0x8000000 ) {
	  l |= 0xf0000000;
	}
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_S18_3BE:
	l >>= 6;
	if ( l & 0x4000000 ) {
	  l |= 0xfc000000;
	}
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_FLOAT_BE:
	p_data = (unsigned char *)&f;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_FLOAT64_BE:
	p_data = (unsigned char *)&d;
	*p_dst++ = p_data[7];
	*p_dst++ = p_data[6];
	*p_dst++ = p_data[5];
	*p_dst++ = p_data[4];
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
	  
      case SND_PCM_FORMAT_U8:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U16_LE:
	l += 0x80000000;
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U16_BE:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	break;
      case SND_PCM_FORMAT_U24_LE:
	l += 0x80000000;
	break;
      case SND_PCM_FORMAT_U24_BE:
	l += 0x80000000;
	break;
      case SND_PCM_FORMAT_U32_LE:
	l += 0x80000000;
	*p_dst++ = p_data[0];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U32_BE:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[0];
	break;
      case SND_PCM_FORMAT_U24_3LE:
	l += 0x80000000;
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U24_3BE:
	l += 0x80000000;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_U20_3LE:
	l += 0x80000000;
	l >>= 4;
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U20_3BE:
	l += 0x80000000;
	l >>= 4;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_U18_3LE:
	l += 0x80000000;
	l >>= 6;
	*p_dst++ = p_data[1];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[3];
	break;
      case SND_PCM_FORMAT_U18_3BE:
	l += 0x80000000;
	l >>= 6;
	*p_dst++ = p_data[3];
	*p_dst++ = p_data[2];
	*p_dst++ = p_data[1];
	break;
      case SND_PCM_FORMAT_IEC958_SUBFRAME_LE:
      case SND_PCM_FORMAT_IEC958_SUBFRAME_BE:
      case SND_PCM_FORMAT_MU_LAW:
      case SND_PCM_FORMAT_A_LAW:
      case SND_PCM_FORMAT_IMA_ADPCM:
      case SND_PCM_FORMAT_MPEG:
      case SND_PCM_FORMAT_GSM:
      case SND_PCM_FORMAT_SPECIAL:
      default:
	fprintf(stderr, "Unsupported output format %d \nSorry.\n",
		pas->output_format);
	exit (1);
      }
      psb->buffer_start += pas->bytes_per_outsample;
      local_buf_len -= pas->bytes_per_insample;
      local_buf_start += pas->bytes_per_insample;
#endif	// DSD22M
    }
  }
 flush_buffer:
  if (result < 0) {
    return 0;
  }
  DEBUG_ASSERT(result % pmf->wfx.Format.nBlockAlign == 0);
  DEBUG_ASSERT(read_count % pmf->wfx.Format.nBlockAlign == 0);
  DEBUG_ASSERT(local_buf_len % pmf->wfx.Format.nBlockAlign == 0);
  DEBUG_ASSERT(local_buf_len < BUFFER_SIZE);
  DEBUG_ASSERT(local_buf_start % pmf->wfx.Format.nBlockAlign == 0);

  if ( psb_prev != NULL && psb_prev->pmf != psb->pmf ) {
    sounddata_closefile(psb_prev->pmf);
  }
  if ( psb->buffer_start > 0 ) {
    psb_prev = psb;
    CYCLIC_INCREMENT(soundbuffer_toread, buffer_num);
    memset(&soundbuffer[soundbuffer_toread], 0, sizeof(SoundBuffer));
    sem_wait(&semaphore);
    pmf->left_buffer_count++;
    soundbuffer_filled++;
    sem_post(&semaphore);
  }

  return (read_count);
}

static void usage(const char *command)
{
  printf("usage: %s -D device_name [-B buffsize] sound_file...\n\n", command);
}

static void interrupt(int signum)
{
  interrupted = TRUE;
}

extern void soundDisplay(MusicFile *pmf);
static void main_loop(void *dummy)
{
  while ( !thread_end_flag && !interrupted ) {
    if ( soundbuffer_filled >= (buffer_num - 1)
	    && p_reading->sound_left_flag ) {
      usleep(WAIT_TIME);
    } else if ( musicfile_reading >= files_open ) {
      usleep(WAIT_TIME);
    } else {
      usleep(0);
    }

    if ( musicfile_reading < files_open && !p_reading->sound_left_flag ) {
      p_reading++;
      musicfile_reading++;
    }
    if ( soundbuffer_filled < (buffer_num - 1) ) {
      if ( musicfile_reading < files_open ) {
        soundDisplay(p_reading);
	read_audio(p_reading);
      } else if ( soundbuffer[soundbuffer_toread].buffer_start > 0) {
        CYCLIC_INCREMENT(soundbuffer_toread, buffer_num);
        soundbuffer[soundbuffer_toread].buffer_start = 0;
        sem_wait(&semaphore);
        soundbuffer_filled++;
        sem_post(&semaphore);
      }
    }
  }
}

int main(const int argc, const char **argv)
{
  char c;
  int i, j, skip, ret;
  size_t memsize;
  pthread_t thread_id;
  //pthread_t thread_id2;
  struct sched_param sp = {50};
  void *dummy;
  char **fnames;

  setlocale(LC_ALL, NULL);
  signal(SIGINT, interrupt);

#ifdef	WITHOUT_DSD
  initFfmpeg();
#endif

  memsize = 0;
  skip = 1;
  for ( i = skip; i < argc; i++ ) {
    if ( i < argc - 1 && strncmp(argv[i], "-D", 2) == 0 ) {
      strncpy(devname, argv[i+1], 1024);
      skip += 2;
      i++;
    } else if ( strncmp(argv[i], "--count", 7) == 0 ) {
      count_mode = TRUE;
      counted_files = 0;
      skip++;
    } else if ( strncmp(argv[i], "--version", 9) == 0 ) {
        printf(VERSION " %s\n",copyright);
	return 0;
    } else if ( i < argc - 1 && strncmp(argv[i], "--skip", 6) == 0 ) {
      if ( sscanf(argv[i + 1], "%d", &skip_sec) != 1 ) {
	skip_sec = 0;
      }
      i++;
      skip += 2;
    } else if ( i < argc - 1 && strncmp(argv[i], "-B", 2) == 0 ) {
      if ( sscanf(argv[i + 1], "%d%c", &j, &c) == 2 ) {
        memsize = j;
	switch (c) {
	case 'K':
	  memsize *= 1024;
	  break;
	case 'M':
	  memsize *= 1048576;
	  break;
	default:
	  if ( sscanf(argv[i + 1], "%d", &j) != 1 ) {
	    memsize = 0;
	  } else {
	    memsize = j;
	  }
	  break;
	}
      } else {
	usage(argv[0]);
	exit(1);
      }
      skip += 2;
      i++;
    } else {
      break;
    }
  }
  if (argc <= skip ) {
    /* sound filename parameter not found */
    usage(argv[0]);
    exit(1);
  }

  int slen = strlen(argv[0]);
  if (slen >= 10) {
    if (strcmp(argv[0] + slen - 10, "dsd22Mplay") == 0) {
      divider = 1;
    } else if (strcmp(argv[0] + slen - 10, "dsd11Mplay") == 0) {
      divider = 2;
    } else if (strcmp(argv[0] + slen - 9, "dsd5Mplay") == 0) {
      divider = 4;
    } else {
      divider = 1;
    }
  }

  fnames = dir2files(skip, argc, argv);
  for (files_open = 0; fnames[files_open] != NULL; files_open++)
    ; // emptyr loop body
  dir2filesReset();

  if ( !count_mode ) {
#ifndef	BULKPLAY
    if ( devname[0] == '\0' ) {
      usage(argv[0]);
      exit (1);
    }
#endif	// BULKPLAY

    /* set process priority */
    if (geteuid() == 0 && sched_setscheduler(0, SCHED_RR, &sp) < 0) {
      perror(NULL);
      fprintf(stderr, "sched_setscheduler() failed.\n");
      exit(1);
    }

#ifndef	WITHOUT_DSD
    sound_get_dsd_type();
    DEBUG_PRINT("DSD type %d\n", e_dsd_type);
#endif

    /* initialize semaphore */
    ret = sem_init(&semaphore, 0, 1);
    if ( ret < 0 ) {
      perror("sem_init() failed.");
      exit(1);
    }
    ret = sem_init(&callback_semaphore, 0, 1);
    if ( ret < 0 ) {
      perror("sem_init() failed.");
      exit(1);
    }
  }

  if ( memsize <= 0 ) {
    memsize = BUFFER_NUM * BUFFER_SIZE; /* reset buffer size */
  }
  if ( memsize < BUFFER_NUM * BUFFER_SIZE ) {
    fprintf(stderr, "buffer size %dBytes is too small. reset to %dBytes.\n",
	    (int)memsize, BUFFER_NUM * BUFFER_SIZE);
    buffer_num = BUFFER_NUM;
  } else {
    buffer_num = memsize / BUFFER_SIZE;
  }
  soundbuffer = malloc(sizeof(SoundBuffer) * buffer_num);
  if ( soundbuffer == NULL ) {
    fprintf(stderr, "malloc() failed.\n");
    exit(1);
  }
  memset(soundbuffer, 0, sizeof(SoundBuffer) * buffer_num);
  musicfile = malloc(sizeof(MusicFile) * files_open);
  if ( musicfile == NULL ) {
    fprintf(stderr, "malloc() failed.\n");
    exit(1);
  }
  memset(musicfile, 0, sizeof(MusicFile) * files_open);

  files_open = 0;
  for ( i = skip; i < argc; i++) {
    fnames = dir2files(i, i + 1, argv);
    for (j = 0; fnames[j] != NULL; j++) {
      musicfile[files_open++].filename = strdup(fnames[j]);
    }
    dir2filesReset();
  }
  dir2filesClear();

  p_reading = musicfile;
  /* pre open all sound file */
  for ( i = 0; i < files_open; i++ ) {
    p_reading->valid_flag = TRUE;
    p_reading->sound_left_flag = TRUE;
    if (count_mode) {
      if ( sounddata_openfile(p_reading) == 0 ) {
        sounddata_closefile(p_reading);
        counted_files++;
      }
    } else {
      if ( sounddata_openfile(p_reading) < 0 ) {
        exit(1);
      }
      sound_format_test(p_reading);
    }
    p_reading++;
  }

  if (count_mode) {
    printf("%d\n", counted_files);
    return (counted_files == 0) ? -1 : 0;
  }

  p_reading = musicfile;
  musicfile_reading = 0;
#ifdef	DSD22M
  coeff44k352k22M_init(2, divider);
#endif
  /* pre reading to fill buffers */
  while ( musicfile_reading < files_open ) {
    while ( p_reading->sound_left_flag ) {
#if 0
      if ( read_audio(p_reading) <= 0 ) {
	p_reading->sound_left_flag = FALSE;
	continue;
      }
#else
      read_audio(p_reading);
#endif
      printf(CR "reading file %d/%d buffer %d/%d", musicfile_reading, files_open, soundbuffer_filled, buffer_num);
      fflush(stdout);
      if ( skip_sec > p_reading->sample_output / p_reading->wfx.Format.nSamplesPerSec ) {
	p_reading->left_buffer_count--;
	p_reading->sample_output += soundbuffer[soundbuffer_towrite].buffer_start/ ( p_reading->wfx.Format.nChannels * snd_pcm_format_physical_width(p_reading->fmt.output_format) / 8);
	CYCLIC_INCREMENT(soundbuffer_towrite, buffer_num);
	soundbuffer_filled--;
      }

      if ( soundbuffer_filled >= (buffer_num - 2) ) {
	goto buffer_full;
      }
    }
    p_reading++;
    musicfile_reading++;
  }
 buffer_full:
  //printf("\r                                          \r");
  fflush(stdout);

  /* initialize sound */
  p_current = NULL;
  thread_end_flag = FALSE;

  if ( pthread_create(&thread_id, NULL, playloop, NULL) != 0 ) {
    fprintf(stderr, "pthread_create() failed.\n");
    exit(1);
  }
#ifdef	NOTDEF
  /* main read loop */
  if ( pthread_create(&thread_id2, NULL, main_loop, NULL) != 0 ) {
    fprintf(stderr, "pthread_create() failed.\n");
    exit(1);
  }
  if ( pthread_join(thread_id2, &dummy) != 0 ) {
    fprintf(stderr, "pthread_join() failed.\n");
    exit(1);
  }
#else
  main_loop(NULL);
#endif

#ifdef	DSD22M
  coeff44k352k22M_finalize();
#endif
  if (interrupted) {
    sem_wait(&semaphore);
    soundbuffer_filled = 0;
    sem_post(&semaphore);
  }

  if ( pthread_join(thread_id, &dummy) != 0 ) {
    fprintf(stderr, "pthread_join() failed.\n");
    exit(1);
  }

  if ( musicfile_reading > 0 ) {
    sounddata_closefile(musicfile + musicfile_reading - 1);
  }

  if ( soundbuffer != NULL ) {
    free(soundbuffer);
    soundbuffer = NULL;
  }
  if ( musicfile != NULL ) {
    for (i = 0; i < files_open; i++) {
      free(musicfile[i].filename);
    }
    free(musicfile);
    musicfile = NULL;
  }

  sem_destroy(&semaphore);
  sem_destroy(&callback_semaphore);

  return 0;
}
