#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#include "usbmydev.h"
#include "common.h"

int usb_getdesc(const char *devfile, struct usb_device_descriptor* desc)
{
	int result = FALSE;
	int fd = open(devfile, O_RDONLY);
	if (fd < 0) {
		perror("");
		return FALSE;
	}
	
	memset(desc, '\0', sizeof(struct usb_device_descriptor));
	ssize_t len = read(fd, desc, sizeof(struct usb_device_descriptor));
	if (len < 0) {
		perror("");
	} else {
		result = TRUE;
	}

	close(fd);

	return result;
}

int usb_open(const char *devfile)
{
	int fd = open(devfile, O_RDWR);
	return fd;
}

int usb_isdevicefree(int fd, int interface)
{
	int ret;

	struct usbdevfs_getdriver driver_info;
	memset(&driver_info, '\0', sizeof(struct usbdevfs_getdriver));
	driver_info.interface = interface;
	ret = ioctl(fd, USBDEVFS_GETDRIVER, &driver_info);
	if (ret < 0 && errno == ENODATA) {
		return TRUE;
	}
	return FALSE;
}

int usb_setinterface(int fd, int interface, int altsetting)
{
	struct usbdevfs_setinterface interface_info;
	interface_info.interface  = interface;
	interface_info.altsetting = altsetting;
	return ioctl(fd, USBDEVFS_SETINTERFACE, &interface_info);
}

int usb_claim(int fd, unsigned int interface)
{
	return ioctl(fd, USBDEVFS_CLAIMINTERFACE, &interface);
}

int usb_release(int fd, unsigned int interface)
{
	return ioctl(fd, USBDEVFS_RELEASEINTERFACE, &interface);
}

int usb_ctrl(int fd, struct usbdevfs_ctrltransfer *ctrl)
{
	return ioctl(fd, USBDEVFS_CONTROL, ctrl);
}

int usb_submiturb(int fd, struct usbdevfs_urb* urbp)
{
	return ioctl(fd, USBDEVFS_SUBMITURB, urbp);
}

int usb_reapurb(int fd, struct usbdevfs_urb* urbpp)
{
	return ioctl(fd, USBDEVFS_REAPURB, urbpp);
}

int usb_reapurb_ndelay(int fd, struct usbdevfs_urb** urbpp)
{
	return ioctl(fd, USBDEVFS_REAPURBNDELAY, (void *)urbpp);
}

int usb_discardurb(int fd, struct usbdevfs_urb* urbp)
{
	return ioctl(fd, USBDEVFS_DISCARDURB, urbp);
}

#define	RAD001BULK_VENDOR	0x1fc9
#define	RAD001BULK_PRODUCT	0x3006

static int is_rad001bulk(char * dev_name)
{
        struct usb_device_descriptor usb_desc;

        if (!usb_getdesc(dev_name, &usb_desc)) {
		return FALSE;
	}
        
        return (usb_desc.idVendor == RAD001BULK_VENDOR && usb_desc.idProduct == RAD001BULK_PRODUCT);
}

#define BASE_DIR_UDEV	"/dev/bus/usb"
#define BASE_DIR_USBFS	"/proc/bus/usb"

const char * rad001_getdevname()
{
	char * top_directory;
	int result = FALSE;
	static char device_name[2048];
	struct stat stat_buf;
	struct dirent *p_entry1, *p_entry2;
	DIR *p_dir1 = NULL, *p_dir2 = NULL;

	top_directory = BASE_DIR_UDEV;

	if (stat(top_directory, &stat_buf) < 0 || !S_ISDIR(stat_buf.st_mode)) {
		top_directory = BASE_DIR_USBFS;
	}

	p_dir1 = opendir(top_directory);
	if (p_dir1 == NULL) {
		goto close_end;
	}
	while ( (p_entry1 = readdir(p_dir1)) != NULL ) {
		if (p_entry1->d_type != DT_DIR || p_entry1->d_name[0] == '.') {
			continue;
		}
		sprintf(device_name, "%s/%s", top_directory, p_entry1->d_name);
		p_dir2 = opendir(device_name);
		if (p_dir2 == NULL) {
			continue;
		}
		while ( (p_entry2 = readdir(p_dir2)) != NULL ) {
			if (p_entry2->d_type != DT_CHR) {
				continue;
			}
			sprintf(device_name, "%s/%s/%s", top_directory, p_entry1->d_name, p_entry2->d_name);
			if (!is_rad001bulk(device_name)) {
				continue;
			}
			result = TRUE;
			goto close_end;
		}
	}
close_end:
	if (p_dir1 != NULL) {
		closedir(p_dir1);
	}
	if (p_dir2 != NULL) {
		closedir(p_dir2);
	}
	return (result ? device_name : NULL);
}
