#include <string.h>
#include <stdio.h>
#include "readwrite.h"
#include "common.h"

extern int myread(int, void *, int);
extern int myseek(int fd, off_t offset, int whence);

/* main file convert function	---------------------------------------------*/
int	readWavHeader(MusicFile *pmf)
{
	int rfd = pmf->fd;
	unsigned char buff[1024];
	unsigned int l, chunkSize;
	unsigned long long samples;
	unsigned long long totalLen;
	enum mode { E_BIT32, E_BIT64 } e_mode;
	int bytesPerSample;

	if ( myread(rfd, buff, 4) != 4 ) {
		return (-1);
	}
	if ( memcmp(buff, "RF64", 4) == 0 ) {
		e_mode = E_BIT64;
	} else if ( memcmp(buff, "RIFF", 4) == 0 ) {
		e_mode = E_BIT32;
	} else {
		return (-1);
	}

	if ( e_mode == E_BIT64 ) {
		/* parse 64 bit RIFF format header */
		if ( !read32bitint(rfd, &l) || l != 0xffffffff ) {
			return (-1);
		}
		if ( myread(rfd, buff, 4) != 4 || memcmp(buff, "WAVE", 4) != 0 ) {
			return (-1);
		}
		if ( myread(rfd, buff, 4) != 4 || memcmp(buff, "ds64", 4) != 0 ) {
			return (-1);
		}
		if ( !read32bitint(rfd, &chunkSize) ) {
			return (-1);
		}
		if ( myread(rfd, buff, 8) != 8 ) {
			return (-1);
		}
		if ( !read64bitlong(rfd, (u_int64_t *)&totalLen) ) {
			return (-1);
		}
		if ( !read64bitlong(rfd, (u_int64_t *)&samples) ) {
			return (-1);
		}
		if ( myseek(rfd, chunkSize - 24, SEEK_CUR) != 0 ) {
			return (-1);
		}
	} else {
		if ( myseek(rfd, 4, SEEK_CUR) != 0 ) {
			return (-1);
		}
		if ( myread(rfd, buff, 4) != 4 || memcmp(buff, "WAVE", 4) != 0 ) {
			return (-1);
		}
	}
	if ( myread(rfd, buff, 4) != 4 ) {
		return (-1);
	}
	while ( memcmp(buff, "fmt ", 4) != 0 ) {
		if ( !read32bitint(rfd, &chunkSize) ) {
			return (-1);
		}
		if ( myseek(rfd, chunkSize, SEEK_CUR) != 0 ) {
			return (-1);
		}
		if ( (chunkSize & 1) == 1 ) {
			if ( myseek(rfd, 1, SEEK_CUR) != 0 ) {
				return (-1);
			}
		}
		if ( myread(rfd, buff, 4) != 4 ) {
			return (-1);
		}
	}

	if ( !read32bitint(rfd, &chunkSize) ) {
		return (-1);
	}
	memset(&pmf->wfx, 0, sizeof(pmf->wfx));
	if ( chunkSize > sizeof(pmf->wfx) ) {
		return (-1);
	}
	if ( myread(rfd, (unsigned char *)&pmf->wfx, chunkSize) != chunkSize ) {
		return (-1);
	}
	if ( myread(rfd, buff, 4) != 4 ) {
		return (-1);
	}
	while ( memcmp(buff, "data", 4) != 0 ) {
		if ( !read32bitint(rfd, &chunkSize) ) {
			return (-1);
		}
		if ( myseek(rfd, chunkSize, SEEK_CUR) != 0 ) {
			return (-1);
		}
		if ( (chunkSize & 1) == 1 ) {
			if ( myseek(rfd, 1, SEEK_CUR) != 0 ) {
				return (-1);
			}
		}
		if ( myread(rfd, buff, 4) != 4 ) {
			return (-1);
		}
	}
	if ( !read32bitint(rfd, &chunkSize) ) {
		return (-1);
	}
	switch ( e_mode ) {
	case E_BIT32:
		pmf->data_len = pmf->total_len = chunkSize / pmf->wfx.Format.nChannels * 2;
		break;
	case E_BIT64:
		if ( totalLen != samples * pmf->wfx.Format.nBlockAlign ) {
			return (-1);
		}
		pmf->data_len = pmf->total_len = totalLen * 2 / pmf->wfx.Format.nChannels;
		break;
	default:
		return (-1);
	}

	bytesPerSample = pmf->wfx.Format.nBlockAlign / pmf->wfx.Format.nChannels;
	pmf->fmt.real_channels = pmf->wfx.Format.nChannels;

	pmf->wfx.Format.nBlockAlign = bytesPerSample * 2;
	pmf->wfx.Format.nChannels = 2;
	return (0);
}

int readWavData(MusicFile *pmf, unsigned char *buff, int len)
{
  unsigned char *read_buff;
  int read_len;
  int bytesPerSample = pmf->wfx.Format.nBlockAlign / pmf->wfx.Format.nChannels;

  read_buff = alloca(bytesPerSample * pmf->fmt.real_channels);
  
  for ( read_len = 0; read_len < len; read_len += bytesPerSample * 2 ) {
    if ( myread(pmf->fd, read_buff, bytesPerSample * pmf->fmt.real_channels) != bytesPerSample * pmf->fmt.real_channels ) {
      break;
    }
    memcpy( buff + read_len, read_buff, bytesPerSample * 2);
  }
  return (read_len);
}
/* end of file	=============================================================*/
