/*
 * wsd.c
 * access WSD file information.
 */
#include <string.h>
#include "readwrite.h"
#include "common.h"

#define BLOCK_SIZE 4096
#define WSD_MAX_CHANNEL 8

extern int myread(int, void *, int);
extern int myseek(int fd, off_t offset, int whence);


int readWsdData(MusicFile *pmf, unsigned char *buff, int len)
{
	unsigned char read_buff[10];
	int read_len;

	len = len / 8 * 8;
	for ( read_len = 0; read_len < len; read_len += 2 ) {
		if ( myread(pmf->fd, read_buff, (size_t)pmf->fmt.real_channels) != pmf->fmt.real_channels ) {
			while (read_len & 7) {
				buff[read_len++] = DSD_ZERO_POINT;
			}
			return (read_len);
		}
		buff[read_len] = read_buff[(int)pmf->fmt.channelMatrix[0]];
		buff[read_len + 1] = read_buff[(int)pmf->fmt.channelMatrix[1]];
	}

	return (read_len);
}

int	readWsdHeader(MusicFile *pmf)
{
  int rfd = pmf->fd;
  unsigned char	buff[4];
  int version;
  unsigned int	text_offset, data_offset, file_size, samplingRate;
  int i;
  int channels = 0;
  unsigned char c;

  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }
  if ( memcmp(buff, "1bit", 4) != 0 ) {
	return (-1);
  }

  /* dummy read 4-7 */
  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }

  myread(rfd, &c, 1);
  version = c;

  /* dummy read 9-11 */
  if ( myread(rfd, buff, 3) != 3 ) {
	return (-1);
  }

  if ( !read32bitBEint(rfd, &file_size) ) {
 	return (-1);
  }

  /* dummy read 16-19 */
  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }

  if ( !read32bitBEint(rfd, &text_offset) ||
	  !read32bitBEint(rfd, &data_offset) ) {
	return (-1);
  }

  /* dummy read 28-31 */
  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }

  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }

  if ( !read32bitBEint(rfd, &samplingRate) ) {
	return (-1);
  }
  /* dummy read 40-43 */
  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }
  myread(rfd, &c, 1);
  channels = c;
  if ( channels > WSD_MAX_CHANNEL ) {
	  return (-1);
  }
  pmf->fmt.real_channels = channels;
  if ( channels == 2 ) {
    pmf->data_len = pmf->total_len = (file_size - data_offset);
  } else {
    pmf->data_len = pmf->total_len = (file_size - data_offset) * 2 / channels;
  }
  /* dummy read 45-47 */
  if ( myread(rfd, buff, 3) != 3 ) {
	return (-1);
  }

  if ( myread(rfd, buff, 4) != 4 ) {
	return (-1);
  }
	   if ( buff[0] == 0 && buff[1] == 0 && buff[2] == 0 && buff[3] == 0 ) {
		   buff[0] = 0x44;	// hack to play broken header file. (xx)
	   }
  if ( (buff[0] & 0x44) == 0x44 ) {
	pmf->fmt.channelMatrix[0] = 0;
	pmf->fmt.channelMatrix[1] = 1;
	if ( buff[0] & 0x20 ) pmf->fmt.channelMatrix[1]++;
	if ( buff[0] & 0x10 ) pmf->fmt.channelMatrix[1]++;
	if ( buff[0] & 0x08 ) pmf->fmt.channelMatrix[1]++;
  } else if ( (buff[0] & 0x28) == 0x28 ) {
	pmf->fmt.channelMatrix[0] = 0;
	pmf->fmt.channelMatrix[1] = 1;
	if ( buff[0] & 0x40 ) {
		pmf->fmt.channelMatrix[0]++;
		pmf->fmt.channelMatrix[1]++;
	}
	if ( buff[0] & 0x10 ) pmf->fmt.channelMatrix[1]++;
	for ( i = 2; i < WSD_MAX_CHANNEL; i++ ) {
		pmf->fmt.channelMatrix[i] = i;
	}
	if ( pmf->fmt.channelMatrix[0] != 0 ) {
		pmf->fmt.channelMatrix[(int)pmf->fmt.channelMatrix[0]] = 0;
	}
	if ( pmf->fmt.channelMatrix[1] != 1 ) {
		pmf->fmt.channelMatrix[(int)pmf->fmt.channelMatrix[1]] = 1;
	}
  } else {
	return (-1);
  }
  if ( myseek(rfd, data_offset, SEEK_SET) != 0 ) {
		return (-1);
  }
  pmf->wfx.Format.nSamplesPerSec = (samplingRate / 32);

  return (0);
}
